#!/usr/bin/env python3
import rospy
import networkx as nx
import std_msgs.msg
from move_base_msgs.msg import MoveBaseGoal
import scipy as sp
import scipy.sparse  # call as sp.sparse
from visualization_msgs.msg import MarkerArray, Marker
import  time

from drive.srv._check import check, checkResponse

graph = nx.Graph()

class graph_metod():
    def __init__(self):
        rospy.init_node('graph_node')
        rospy.Subscriber('pose_for_graph', MoveBaseGoal, self.goal_callback)
        rospy.Subscriber('spaces_for_graph', std_msgs.msg.Int32MultiArray, self.space_storage)
        rospy.Subscriber('chosen_space', std_msgs.msg.Int32MultiArray, self.choise_storage)
        self.check_pub = rospy.Publisher('goal_check', std_msgs.msg.Bool, queue_size=10)
        self.marker_pub = rospy.Publisher('visualization_marker_array', MarkerArray, queue_size=10)
        self.node_counter = 0
        self.choise_node = 0
        self.result = std_msgs.msg.Bool()
        graph.add_node(self.node_counter)
        self.node_counter += 1

        self.marker_array = MarkerArray()
        self.goal_marker = Marker()
        self.goal_marker.type = Marker.CYLINDER
        self.goal_marker.header.frame_id = 'map'
        self.goal_marker.color.r = 1.0
        self.goal_marker.color.g = 2.0
        self.goal_marker.color.b = 0.0
        self.goal_marker.color.a = 0.2
        self.goal_marker.scale.z = 0.1
        self.goal_marker.scale.x = 1.5
        self.goal_marker.scale.y = 1.5
        self.marker_counter = 0
        self.service = rospy.Service('goal_check', check, self.handle)

        self.r = 1.5

        # self.check_pub = rospy.Publisher('space_check', std_msgs.msg.Bool, queue_size=10)
        self.rate = rospy.Rate(10)

    def handle(self, req):
        for node, data in graph.nodes(data='goal'):
            # print(node)
            # print(type(data))
            # print(type(req))
            if data is not None:
                try:
                    if (req.goal.target_pose.pose.position.x - data.target_pose.pose.position.x)**2 + (req.goal.target_pose.pose.position.y - data.target_pose.pose.position.x)**2 < self.r**2:
                        print('comparing positive')
                        return checkResponse(True)
                except Exception as error:
                    print(error)
                    return checkResponse(False)
        print('comparing negative')
        self.goal_marker.pose.position.x = req.goal.target_pose.pose.position.x
        self.goal_marker.pose.position.y = req.goal.target_pose.pose.position.y
        self.marker_array.markers.insert(self.marker_counter ,self.goal_marker)
        self.marker_pub.publish(self.marker_array)
        self.marker_counter += 1
        return checkResponse(False)

    def incidence_matrix(G, nodelist=None, edgelist=None, oriented=False, weight=None):
        
        import scipy as sp
        import scipy.sparse  # call as sp.sparse

        if nodelist is None:
            nodelist = list(G)
        if edgelist is None:
            if G.is_multigraph():
                edgelist = list(G.edges(keys=True))
            else:
                edgelist = list(G.edges())
        A = sp.sparse.lil_array((len(nodelist), len(edgelist)))
        node_index = {node: i for i, node in enumerate(nodelist)}
        for ei, e in enumerate(edgelist):
            (u, v) = e[:2]
            if u == v:
                continue  # self loops give zero column
            try:
                ui = node_index[u]
                vi = node_index[v]
            except KeyError as err:
                raise nx.NetworkXError(
                    f"node {u} or {v} in edgelist but not in nodelist"
                ) from err
            if weight is None:
                wt = 1
            else:
                if G.is_multigraph():
                    ekey = e[2]
                    wt = G[u][v][ekey].get(weight, 1)
                else:
                    wt = G[u][v].get(weight, 1)
            if oriented:
                A[ui, ei] = -wt
                A[vi, ei] = wt
            else:
                A[ui, ei] = wt
                A[vi, ei] = wt
        return A.asformat("csc")

    def choise_storage(self ,msg): 
        self.choise = msg.data
        # print('choise recieved - graph')
        # print('choise - ', self.choise)
    
    def space_storage(self ,msg):          # reminder for tomorrow - spaces update to fast, they need to be stored in space_storage until goal is reached, then aplied to graph
                                      # reminder for tomorrow - now, in drive.py, chosen space need to be send here, and be starting node for future tracking (eg. if robot chose space with index 2, then 
                                    # next node should be 2 and so on
        self.spaces = []                     #reminder for tomorrow - chosen space is being send here, continue with next part of previous reminder
        for i in msg.data:
            self.spaces.append(i)
        self.grouped_spaces = [self.spaces[n:n+2] for n in range(0, len(self.spaces), 2)]
        # print('space recieved - graph')
        # print('grouped spaces - ',grouped_spaces)
        # print('from graph_node - spaces - ', self.spaces)                         #plan for tomorrow - need to create a function, that will check if next goal will be set in general area where 
                                                                                    #robot has already been. to do this, use odom data - check, if present odom + vector of the next goal will be 
                                                                                    #in general area of previous odom data
            
    def goal_callback(self ,msg):                                                   
        # print(msg)
        self.goal = msg
        # print('goal recieved - graph')

    def spin(self):
        print('starting graph spin')
        while True:
            try:
                self.grouped_spaces
                self.choise
                self.goal
                print('waiting for variables')
            except:
                continue
            else:
                break
        # self.result.data = self.goal_check(self.goal, 2)
        # while self.check_pub.get_num_connections() == 0:
        #         print('connections - ', self.check_pub.get_num_connections())
        #         self.rate.sleep()
        # if self.result.data == True:
        #     while self.check_pub.get_num_connections() == 0:
        #         # print('sleeeeeeeeeepyyyy')
        #         self.rate.sleep()
        #     self.check_pub.publish(self.result)
        #     del self.grouped_spaces 
        #     del self.choise 
        #     del self.goal
        #     self.spin()
        # else:
        #     time.sleep(2)
            # print(type(self.result))
            # print(self.result)
            # print(self.check_pub.get_num_connections())
            # while self.check_pub.get_num_connections() == 0:
            #     print('sleeeeeeeeeepyyyy')
            #     self.rate.sleep()
            # self.check_pub.publish(self.result)
        graph.add_node(self.node_counter)
        nx.set_node_attributes(graph, {self.node_counter:self.goal}, 'goal')
        # buff = self.node_counter
        # print(type(self.grouped_spaces))                                        #reminder - self.grouped_spaces not being created, probably because of send_goal function in drive.py,
                                                                                #possible fix - make spin here react to goal_check result accordingly, while also stop send_goal from sending goal here
        for i in self.grouped_spaces:
            # print('edge_added')
            # print(self.choise[0], i[0])
            # print(self.choise[1], i[1])
            if self.node_counter == self.choise_node:
                self.node_counter += 1
            graph.add_edge(self.choise_node, self.node_counter, space=i)
            self.node_counter += 1
        self.choise_node += 1
        # for node, data in graph.nodes(data='goal'):
        #     try:
        #         print(node)
        #         print(data.target_pose.pose.position.x)
        #         # print(graph.nodes[self.choise_node]['goal'].target_pose.pose.position)
        #     except Exception as error:
        #         print(error)
        # print(graph.nodes(data=True))
        self.previous_choise = self.choise
        # node_counter = buff
        # print('from graph_node - edge data - ', graph.edges.data('space'))
        # node_counter += 1
        print('edges --', graph.edges)
        # print('nodes --', graph.nodes)
        # print('incidence_matrix')
        # print(incidence_matrix(graph))
        del self.grouped_spaces 
        del self.choise 
        del self.goal


if __name__ == '__main__':
    Graph = graph_metod()
    while not rospy.is_shutdown():
        # print('spining')
        Graph.spin()
