#!/usr/bin/env python3
import rospy
import move_base_msgs.msg
import actionlib
from sensor_msgs.msg import LaserScan
from nav_msgs.msg import Odometry
import numpy as np
import math
import time
import std_msgs.msg
import traceback
from drive.srv._check import check, checkRequest

class Move():
    def __init__(self):
        self.node = rospy.init_node('move_node')
        # rospy.Subscriber('goal_check', std_msgs.msg.Bool, self.check_answer)
        self.goal_pub = rospy.Publisher('pose_for_graph', move_base_msgs.msg.MoveBaseGoal, queue_size= 10)
        self.spaces_pub = rospy.Publisher('spaces_for_graph', std_msgs.msg.Int32MultiArray, queue_size= 10)
        self.move_client = actionlib.SimpleActionClient('move_base', move_base_msgs.msg.MoveBaseAction)
        self.choise_pub= rospy.Publisher('chosen_space', std_msgs.msg.Int32MultiArray, queue_size=10)
        self.odom_pub = rospy.Publisher('odom_for_graph', Odometry, queue_size=10)
        self.move_client.wait_for_server()
        self.goal = move_base_msgs.msg.MoveBaseGoal()
        self.goal.target_pose.header.frame_id = 'map'
        self.goal.target_pose.header.stamp = rospy.Time.now()
        self.empty_spaces = std_msgs.msg.Int32MultiArray()
        self.position = 1
        self.chosen_space = std_msgs.msg.Int32MultiArray()
        self.x = 0
        self.y = 0  
        self.odom_x = 0
        self.odom_y = 0
        self.first = True
        self.c = 0.5
        self.key_360 = False

        # self.get_odom = False

    def range_calibration(self, chosen_space, calculated_range):
        lds_ranges = rospy.wait_for_message('/scan_filtered', LaserScan)
        print('chosen ranges - ',chosen_space.data[0], ' - ', lds_ranges.ranges[chosen_space.data[0] - 1], chosen_space.data[1], ' - ', lds_ranges.ranges[chosen_space.data[1] + 1] )
        print('calculated range - ', calculated_range)

    def odom_callback(self):
        msg = rospy.wait_for_message('odom_filter', Odometry)
        self.odom_x = msg.pose.pose.position.x
        self.odom_y = msg.pose.pose.position.y

    def send_goal(self):
        self.goal.target_pose.pose.position.x = self.x
        self.goal.target_pose.pose.position.y = self.y
        self.goal.target_pose.pose.orientation.w = 1.0
        self.move_client.send_goal(self.goal)
        # self.choise_pub.publish(self.chosen_space)
        # self.spaces_pub.publish(self.empty_spaces)

        self.move_client.wait_for_result()

        print('goal achieved')

        # self.goal_pub.publish(self.goal)

        self.empty_spaces.data = []
        
        # self.scan_callback(rospy.wait_for_message('/scan_filtered', LaserScan))

    def goal_check(self):           
        self.goal.target_pose.pose.position.x = self.x
        self.goal.target_pose.pose.position.y = self.y
        self.goal.target_pose.pose.orientation.w = 1.0
        self.spaces_pub.publish(self.empty_spaces)
        self.choise_pub.publish(self.chosen_space)
        self.goal_pub.publish(self.goal)
        rospy.wait_for_service('goal_check')
        try:
            add_two_ints = rospy.ServiceProxy('goal_check', check)
            resp1 = add_two_ints(self.goal)
            return resp1.result
        except rospy.ServiceException as e:
            print("Service call failed: %s"%e)
        # print('check finished')
    
    def vector_calculation(self, ranges):
        try:
            print(self.chosen_space.data[1], self.chosen_space.data[0])

            self.odom_callback()

            print(ranges[self.chosen_space.data[0] - 1], ranges[self.chosen_space.data[1] + 1])
            print('cos - ', math.cos((math.radians(self.chosen_space.data[0] + self.chosen_space.data[1])/2)))

            # print(self.key_360)

            if self.chosen_space.data[0] > self.chosen_space.data[1]:
                self.cos = math.cos((math.radians(self.chosen_space.data[1] - self.chosen_space.data[0])))
            else:
                self.cos = math.cos((math.radians(360 - self.chosen_space.data[1] - self.chosen_space.data[0])))

            self.verctor_lengh = self.c * math.sqrt(float(ranges[self.chosen_space.data[0] - 1]**2 + ranges[self.chosen_space.data[1] + 1]**2 + 2 
                                               * ranges[self.chosen_space.data[1] + 1] * ranges[self.chosen_space.data[0] - 1])) * abs(self.cos)
            if self.cos < 0:
                self.verctor_lengh = self.verctor_lengh * -1
            self.x = ((self.odom_x + self.verctor_lengh)  * math.cos((math.radians((self.chosen_space.data[1] + self.chosen_space.data[0])/2))))
            self.y = ((self.odom_y + self.verctor_lengh)  * math.sin((math.radians((self.chosen_space.data[1] + self.chosen_space.data[0])/2))))
            # print(self.odom_y, self.odom_x, verctor_lengh)
            print("x --" ,self.x , "y -- ", self.y)
            
            # print(self.previous_x, self.previous_y)
        except Exception as error:
            print('Error in vector_calculation', traceback.format_exc())
        # print('vector finished')

    def scan_callback(self, msg):
        self.ranges = np.array(msg.ranges, float)
        indexes = np.where(self.ranges == math.inf)
        print(indexes)
        for i in range(len(indexes[0])):
            try:
                if indexes[0][i + 1] - indexes[0][i] > 30:
                    # print('space found!')
                    self.empty_spaces.data.append(indexes[0][i + 1])
                    self.empty_spaces.data.append(indexes[0][i])

            except:
                break
        try:
            if indexes[0][0] != 0 and indexes[0][-1] != 360:
                # print('space found!')
                self.empty_spaces.data.append(indexes[0][0])
                self.empty_spaces.data.append(indexes[0][-1])

        except Exception as error:
            print(error)
        counter = 0
        self.grouped_spaces = [self.empty_spaces.data[n:n+2] for n in range(0, len(self.empty_spaces.data), 2)]
        print(self.empty_spaces)

        # if self.grouped_spaces[counter][0] < self.grouped_spaces[counter][1]:
        #     temp = self.grouped_spaces[counter][0]
        #     self.grouped_spaces[counter][0] = self.grouped_spaces[counter][1]
        #     self.grouped_spaces[counter][1] = temp

        self.chosen_space.data = self.grouped_spaces[counter]
        print(self.chosen_space.data)

        if len(self.grouped_spaces) < 1:
            self.empty_spaces.data.insert(0, self.chosen_space.data[0])
            self.empty_spaces.data.insert(0, self.chosen_space.data[1])
        # print('scan finished')

    def spin(self):
        print('spin started - drive')
        counter = 0
        self.scan_callback(rospy.wait_for_message('/scan_filtered', LaserScan))
        self.vector_calculation(self.ranges)
        self.range_calibration(self.chosen_space, self.verctor_lengh)
        result = self.goal_check()

        while True:

            if result == True:
                try:
                    counter += 1
                    self.chosen_space.data = self.grouped_spaces[counter]       #for some reason code is stuttering - like, at some point it just stops compleatly, full debug is requared
                    self.vector_calculation(self.ranges)
                    result = self.goal_check()
                    print(counter)
                    print('result -',result)
                    continue
                except:
                    continue
            elif result == False:
                self.send_goal()
                break

if __name__ == "__main__":
    move = Move()
    while not rospy.is_shutdown():
        # print('spining')
        move.spin()